/*
	Copyright (C) 2006 Ivo van Doorn

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Laptop keys support

	INTRODUCTION
	Laptops are quite often equiped with additional hardware keys.
	This includes the radio keys for wireless and bluetooth devices,
	but also other (programmable) keys.
	This driver will create a generic layer between hardware and
	userspace. Drivers can add keys to the input device for
	frequent polling and reporting the status to userspace.

	INPUT DEVICES
	When loading this driver the main input device is directly created.
	When desired drivers can also register a new laptop_key_device
	which creates a new input device to which the laptop keys can be
	registered.

	POLLING
	In the laptop_key_device structure the delay is set, this delay
	is used to determine the polling interval in which each key will be
	polled for its status. When the key_poll field is not set, the
	key will not be polled.

	STATUS REPORTING
	When the status has changed, an input event will be send to userspace.
	The key field will be used as reported key in that input event.
	When the status of the key has changed, this event will also be send to
	the driver that registered the key when the key_event field has been
	set. It is also possible for keys to skip the polling and inform
	this driver of the event directly.
 */

#ifndef LAPTOPKEY_H
#define LAPTOPKEY_H

#include <linux/input.h>
#include <linux/list.h>
#include <linux/spinlock.h>

/**
 * struct laptop_key_device - laptop key device structure
 * @dev_name: Name of the input device.
 * @delay: When polling is required for any of the keys,
 *	this polling delay (in ms) will be used.
 * @list: All laptop keys are part of a linked list.
 *	This field should not be touched by any driver.
 * @input_dev: This is the input device created for this laptop_key_device.
 *	This field should not be touched by any driver.
 * @timer: The timer structure for the frequent polling of the keys.
 *	This field should not be touched by any driver.
 * @lock: lock to protect the list from concurrent access.
 *	This field should not be touched by any driver.
 */
struct laptop_key_device {
	const char *dev_name;

	unsigned int delay;

	struct list_head list;

	struct input_dev *input_dev;

	struct timer_list timer;

	spinlock_t lock;
};

/**
 * struct laptop_key - laptop key structure
 * @key: Key type that is send to userspace.
 *	should be any of the KEY_* defines from <linux/input.h>
 * @data: private data that is send as argument with the
 *	functions key_poll and key_event.
 * @current_status: Current status of the button is stored here.
 * @key_poll: This function is used to determine the current state
 *	of the key. This field only needs to be set when polling is required.
 * @key_event: This function is used when the key status has changed, it will
 *	then send the new key status as argument to inform the hardware
 *	of the change. This field is only required when the hardware needs to
 *	be informed of the change.
 * @keydev: laptop_key_device structure this key is attached to.
 *	This field should not be touched by any driver.
 * @entry: All laptop keys are part of a linked list. This field should not
 *	be touched by any driver.
 */
struct laptop_key {
	unsigned int key;
	unsigned long data;
	int current_status;

	int (*key_poll)(unsigned long data);
	void (*key_event)(unsigned long data, int status);

	struct laptop_key_device *keydev;
	struct list_head entry;
};

/*
 * Functions when using the main input device.
 */
void laptopkey_add_key(struct laptop_key *);
void laptopkey_del_key(struct laptop_key *);

/*
 * Functions when using a seperate input device.
 */
int laptopkey_register_device(struct laptop_key_device *);
void laptopkey_unregister_device(struct laptop_key_device *);
void laptopkey_dev_add_key(struct laptop_key_device *, struct laptop_key *);
void laptopkey_dev_del_key(struct laptop_key_device *, struct laptop_key *);

/*
 * Manually report a key event.
 */
void laptopkey_key_event(struct laptop_key *, int status);

#endif /* LAPTOPKEY_H */
